#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 * ------------------------------------------------------------------------
 * bmSearch
 * ------------------------------------------------------------------------
 * @license : GNU Affero General Public License Version 3 <http://www.gnu.org/licenses>
 * @author  : Pascal SALAUN 
 * @Website : http://www.bm-stats.org <http://www.bm-stats.org>
 * ------------------------------------------------------------------------
"""

# Native Python Lib/Class 
import base64
from collections import defaultdict
import cStringIO
from datetime import *
import magic
import os
import quopri
import re
import subprocess
import sys
import tarfile
import time 


# bmSearch Class
from bmDate import *
import config
from fileMIMEType import *
from regex import *
from requestToELK import *
from sqlQuery import *

io_method = cStringIO.StringIO


class mailLogParser:
        """
        mailLogParser :
                Has to open/stream mail.log file,
                parse each line to feed a dict with pertinent values
                and try (if kwargs set) to publish them in ELK platform
        """

        def __init__(self):
                """
                __init__ : 
                        constructor accept kwargs
                        
                        **kwargs
                                                file   : path to file to parse. I can be plain text, gzip or bzip2 compressed. By default what configured in /etc/bmSearch/config.py
                                                
                                                method : how to read the file, 2 possibilities
                                                        - "file" : read the file and exit after reach end file. By default.
                                                        - "stream" : connect to the file and stream data. No exit while programm running
                                                        
                                                action : tell what to do, 2 cases
                                                        - "storeToELK"  : to store in Elasticsearch instance configured in /etc/bmSearch/config.py
                                                        - "updateToELK" : to update in Elasticsearch instance data (deliveredRecip)
                        
                """

                # Main data dict
                self.data                 = defaultdict(dict)
                self.surveyQueuid         = []
                self.queueidFromMessageid = defaultdict(dict) # Only for Antivirus case
		self.addonInfo            = defaultdict(dict)

		# Update data dict
                self.updateData                 = defaultdict(dict)
                self.updateSurveyQueuid         = []
                
                # Set a tuple of file types to init (cf connexion)
                self.theseFileTypes = config.fileTypeTupl


                """
                Postfix : these lines are about postfix activity for one mail  
                """  
                
                self.smtpdLine       = smtpdLine
                self.cleanupLine     = cleanupLine
                self.filenameLine    = filenameLine
                self.bodyDateLine    = bodyDateLine
                self.bodyFromLine    = bodyFromLine
                self.bodySubjectLine = bodySubjectLine
                self.qmgrLine        = qmgrLine
                self.smtpLineDSN2    = smtpLineDSN2
                self.smtpLineDSN4    = smtpLineDSN4
                self.smtpLineDSN5    = smtpLineDSN5
                self.smtpRemoved     = smtpRemoved

                """
                Cyrus/IMAP : this line is relative to a délivery success
                """
                self.cyrusDelivered = cyrusDelivered

                
                """
                DATE
                """
                self.dates = bmDate().getDates()
                
                
                """
                cf https://stackoverflow.com/questions/7331351/python-email-header-decoding-utf-8 Abhishek Gurjar
                """
                self.pat2=re.compile(r'(([^=]*)=\?([^\?]*)\?([BbQq])\?([^\?]*)\?=([^=]*))',re.IGNORECASE)
        
                
        def setParam(self, **kwargs):
                """
                setParam:
                        overload params with these in **kwargs.
                        Params available are :
                            file   
                            action
                            method
                            dbStore
                            dbtemp
                            dbStoreMaxDays
                            timeInterval
                        For details have a look in /etc/bmSearch/config.py
                """
                self.maillog = config.maillog
                if "file" in kwargs.keys():
                        self.maillog = kwargs['file']

                self.action = ""
                if "action" in kwargs.keys():
                        self.action = kwargs['action']
                         
                self.method = "file"
                if "method" in kwargs.keys():
                        self.method = kwargs['method']
                        
                self.dbStore = config.dbStore
		if "dbStore" in kwargs.keys():
                        self.dbStore = kwargs['dbStore']
                
		self.dbtemp = config.dbtemp
		if "dbtemp" in kwargs.keys():
			self.dbStore = kwargs['dbtemp']

        
                self.dbStoreMaxDays = config.dbStoreMaxDays
                if "dbStoreMaxDays" in kwargs.keys():
                        self.dbStoreMaxDays = kwargs['dbStoreMaxDays']
                        
                        
		self.timeInterval = config.timeInterval
		if "timeInterval" in kwargs.keys():
			self.timeInterval = kwargs['timeInterval']



        def getStreamContent(self):
                """
                getStreamContent :
                        Connect to a file and strema its content (as a tail). Its name is from config or  bmSearch (main) kwargs "file" (see bmSearch).
                        Set var "content" with line(s)
                        Break only when main breaks
                """
                try:
			stream = open(self.maillog)
		        
		        interval = 1.0

		        while True:
		                where = stream.tell()
		                line = stream.readline()
		                
		                if not line:
		                        time.sleep(interval)
		                        stream.seek(where)
		                else:
		                        yield line

        	except Exception as error:
    			self.logger.catch('mailLogParser::getStreamContent')
              
              
              
        def getFileContent(self):
                """
                getFileContent :
                        Read a whole file.  Its name is from config or  bmSearch (main) kwargs "file" (see bmSearch).
                        Set var "content" with line(s)
                
                """
                
		self.maillog    = self.maillog.encode('utf8')

                self.compressionCase = {"text/plain": "plaintext","application/x-gzip" : "r:gz", "application/gzip" : "r:gz", "application/x-bzip2" : "r:bz2" }
                t = fileMIMEType()
                self.ftype =ftype = t.getType(self.maillog)
                
                if self.compressionCase[self.ftype] == "plaintext":
                	try:
                        	f = open(self.maillog , 'r' )
                        	content = f.readlines()        
                        	f.close()
                        except Exception as error:
    				self.logger.catch('mailLogParser::getFileContent::plaintext case')  
                else:
                        try:
                                ft = tarfile.open(self.maillog, self.compressionCase[ftype])
                                member0 = ft.getnames()[0]
                                f=ft.extractfile(member0)
                                content = f.readlines()
                                ft.close()
                                
                        except tarfile.ReadError:
                                p = subprocess.Popen(["zcat", self.maillog], stdout = subprocess.PIPE)
                                content = io_method(p.communicate()[0])
                                assert p.returncode == 0
                                
                        except Exception as error:
    				self.logger.catch('mailLogParser::getFileContent::no compression format found')
                        
                for line in content:
                	yield line
                
        

	def parseContent(self):
                """
                parseContent :
                        read line from getFileContent (=> method = 'file') or getStreamContent (=> method ='stream') and extract data
                        If  bmSearch (main) kwargs "action" is set and has value :
                                                - "storeToELK", try to post data to an Elasticsearch instance
                                                - "updateToELK", try to update an Elasticsearch instance data (deliveredRecip)
                """


		"""
                Open SQLITE DB
                """
		if self.method == "file":
			db = self.dbtemp
		if self.method == "stream":
			db = self.dbStore	

                self.cDb = sqlQuery(db)
		self.cDb.initDb()
                self.cDb.updateTrigger(self.dbStoreMaxDays)


          
                if self.method == "file":
			for line in self.getFileContent():
				self.parseLine(line)

                if self.method == "stream":
			for line in self.getStreamContent():
				self.parseLine(line)



        def decodev2(self, bodySubject):
                """
                decodev2:
                    return email subject in pretty view
                    Args :
                        bodySubject ; the email subject entry
                """
                data=self.pat2.findall(bodySubject)
                line=[]
                if data:
                        for g in data:
                                (raw,extra1,encoding,method,string,extra)=g
                                extra1=extra1.replace('\r','').replace('\n','').strip()
                                if len(extra1)>0:
                                        line.append(extra1)
                                if method.lower()=='q':
                                        string=quopri.decodestring(string)
                                        string=string.replace("_"," ").strip()
                                if method.lower()=='b':
                                        string=base64.b64decode(string)
                                line.append(string.decode(encoding,errors='ignore'))
                                extra=extra.replace('\r','').replace('\n','').strip()
                                if len(extra)>0:
                                        line.append(extra)
                        return "".join(line)
                else:
                        return bodySubject
                    

        def parseLine(self, line):
                """
                parseLine :
                        read line  and extract data
                        If  bmSearch (main) kwargs "action" is set and has value :
                                                - "storeToELK", try to post data to an Elasticsearch instance
                                                - "updateToELK", try to update an Elasticsearch instance data (deliveredRecip)
                """
                thisLine = line.rstrip()

                # Connection            
                smtpdLineEntries      = self.smtpdLine.match(thisLine)
                if smtpdLineEntries is not None:
                       	# try to update some datas (antivirus, cyrus message delivery status)
                
                        date = time.strftime('%Y-%m-%d',time.localtime())
                        try:
                                date = self.dates[smtpdLineEntries.group('date')]

                        except:
                                pass                        
                       
                        hour    = smtpdLineEntries.group('hour')
                        timestp = int(datetime.strptime(date + " " + hour, '%Y-%m-%d %H:%M:%S').strftime("%s")) * 1000
                        queueid = smtpdLineEntries.group('queueid')
                        server  = smtpdLineEntries.group('server').split(",")[0].replace("["," : ").replace("]","")
                        
                        self.data[queueid] = defaultdict(dict)
                        self.data[queueid]['@timestamp'] = timestp
			self.data[queueid]['date'] = date
			self.data[queueid]['hour'] = hour
                        self.data[queueid]['server'] = server
                        
                        # Elastic Entries
                        self.data[queueid]['source'] = self.maillog
                        self.data[queueid]['type'] = 'message'
                        
                        self.data[queueid]['beat'] = defaultdict(dict)
                        self.data[queueid]['beat']['hostname'] = config.HOSTNAME
                        self.data[queueid]['beat']['name']     = 'bmsearch'
                        self.data[queueid]['beat']['version']  = config.VERSION
                        
			self.data[queueid]['from'] = self.data[queueid]['fromDomain'] = self.data[queueid]['nrcpt'] = self.data[queueid]['size'] = ""
			self.data[queueid]['bodyDate'] = self.data[queueid]['bodyFrom'] = self.data[queueid]['bodyFromDomain'] = self.data[queueid]['bodySubject'] = ""
			
			self.data[queueid]['antivirus'] = 'CLEAN : not analyzed !'
		
			self.data[queueid]['filename'] = []	
			self.data[queueid]['recip'] = []
			self.data[queueid]['deferredRecip'] =  []
			self.data[queueid]['bouncedRecip'] = []	
			self.data[queueid]['recipType'] =  defaultdict(dict)
			self.data[queueid]['recipType']['internal'] = 0
			self.data[queueid]['recipType']['external'] = 0
			
			self.data[queueid]['fileType'] =  defaultdict(dict)
                        for thisType in self.theseFileTypes:
                                self.data[queueid]['fileType'][thisType] = 0
                       
                        
                        
                # Message-id         
                cleanupLineEntries    = self.cleanupLine.match(thisLine)        
                if cleanupLineEntries is not None:
                        queueid   = cleanupLineEntries.group('queueid')
                        messageid = cleanupLineEntries.group('messageid').replace("<","").replace(">","")
                        self.data[queueid]['msgid'] = messageid

			# A hack because in case of delivery notification (from=<>), it could not have an "postfix/smtpd" line 
			if 'recip' in self.data[queueid].keys():
                     		pass
			else:
        	                self.data[queueid]['recip'] = []
                        	self.data[queueid]['recipType'] =  defaultdict(dict)
                        	self.data[queueid]['recipType']['internal'] = 0
                        	self.data[queueid]['recipType']['external'] = 0


 
                        if self.cDb.verifyIfMessageidExists(messageid):
                        	# Because we have already parsed this message lines
                        	del self.data[queueid]
                        	
                        	# We are now in update mode, to only catch DSN case and do an update of the existing entry
                                self.updateData[queueid] = defaultdict(dict)
				self.updateData[queueid]['deferredRecip'] = [] 
  				self.updateData[queueid]['bouncedRecip'] = []
				self.updateData[queueid]['msgid'] = messageid
                                self.updateSurveyQueuid.append(queueid)
                                
                        else:
                                self.surveyQueuid.append(queueid)
				self.queueidFromMessageid[messageid] = queueid
				
                                if '@timestamp' in self.data[queueid].keys():
                                        timestamp = self.data[queueid]['@timestamp']/1000
                                else:
                                        timestamp = int(time.time())

                                # We insert messageid an timestamp in sqlite for later, in case we have 2 message blocs
                                tupl=(messageid,timestamp)
                                self.cDb.addEntry(tupl)

                                          
                ## ANTIVIRUS 
                
                # Kasperky Labs
                klmsLineEntries = klmsLine.match(thisLine.replace("\"",""))
                if klmsLineEntries is not None:                        
                        messageid = klmsLineEntries.group('messageid')
                        status    = klmsLineEntries.group('status')
                        status   += ", antivirus-status:" + klmsLineEntries.group('avstatus')
                        status   += ", antiphishing-status:" + klmsLineEntries.group('apstatus')
                        status   += ", antispam-status:" + klmsLineEntries.group('asstatus')
               		
               		if messageid in self.queueidFromMessageid.keys():
               			queueid = self.queueidFromMessageid[messageid]
               			self.data[queueid]['antivirus'] = status
               			try:
               				del self.queueidFromMessageid[messageid]
               			except KeyError:
               				pass
                   
                   
               	
               	#Amavis-New
               	amavisLineEntries = amavisLine.match(thisLine)
                if amavisLineEntries is not None:
                        messageid = amavisLineEntries.group('messageid')
                        status    = amavisLineEntries.group('status')
                        
               		if messageid in self.queueidFromMessageid.keys():
               			queueid = self.queueidFromMessageid[messageid]
               			self.data[queueid]['antivirus'] = status
               			try:
               				del self.queueidFromMessageid[messageid]
               			except KeyError:
               				pass
                       	
                                      
                
                # Filename if controled 
                filenameLineEntries   = self.filenameLine.match(thisLine)
                if filenameLineEntries is not None:
                        queueid  = filenameLineEntries.group('queueid')
                        filename = filenameLineEntries.group('filename')
                        
                        
                        if queueid in self.surveyQueuid:
                                
                                fileType = filename.lower().split('.')[-1].replace("jpg","jpeg")
				
				if filename not in self.data[queueid]['filename']:
					self.data[queueid]['filename'].append(filename)                                
                                	try:
                                        	self.data[queueid]['fileType'][fileType] += 1
                                	except:
                                        	self.data[queueid]['fileType'][fileType] = 0
                                        	self.data[queueid]['fileType'][fileType] += 1
                                        
                                
                                        
                        
                ## Specific body message checks (see /etc/postfix/BMSEARCH_header_checks)
                
                # For these header controls we have to set the self.data[queueid] in every case, because lines
                # are written by cleanup, before the message-id
                # bodyDate if controled 
                bodyDateLineEntries   = self.bodyDateLine.match(thisLine)
                if bodyDateLineEntries is not None:
                        queueid  = bodyDateLineEntries.group('queueid')
                        bodyDate = bodyDateLineEntries.group('bodyDate')
                        
                        self.data[queueid]['bodyDate'] = bodyDate
                                 
                                 
                                        
                # bodyFrom if controled 
                bodyFromLineEntries   = self.bodyFromLine.match(thisLine)
                if bodyFromLineEntries is not None:
                        queueid  = bodyFromLineEntries.group('queueid')
                        bodyFrom = bodyFromLineEntries.group('bodyFrom')
                        
                        self.data[queueid]['bodyFrom'] = bodyFrom.lower().split('<')[-1].replace(">","")
                        self.data[queueid]['bodyFromDomain'] = bodyFrom.lower().split('@')[-1].replace(">","")
                                
                                
                                        
                # bodySubject if controled 
                bodySubjectLineEntries   = self.bodySubjectLine.match(thisLine)
                if bodySubjectLineEntries is not None:
                        queueid     = bodySubjectLineEntries.group('queueid')
                        bodySubject = bodySubjectLineEntries.group('bodySubject')
                        
                        self.data[queueid]['bodySubject'] = self.decodev2(bodySubject)
                                                         
                
                
                # Return to general info.
                # Sender, size, nrcpt        
                qmgrLineEntries       = self.qmgrLine.match(thisLine)
                if qmgrLineEntries is not None:
                        queueid  = qmgrLineEntries.group('queueid')
                        thisFrom = qmgrLineEntries.group('from')
                        size     = qmgrLineEntries.group('size')
                        nrcpt    = qmgrLineEntries.group('nrcpt')

			if len(thisFrom):
				pass
			else:
				thisFrom='notification@none' 

	                        
			if queueid in self.surveyQueuid:
		        	self.data[queueid]['size']   =  size
	        	        self.data[queueid]['nrcpt']  =  nrcpt
	             	        self.data[queueid]['from']   =  thisFrom.lower() 
                               	self.data[queueid]['fromDomain'] = thisFrom.lower().split('@')[-1].replace(">","")
                       

				
 
                        
                # When delivery is OK      
                smtpLineDSN2Entries   = self.smtpLineDSN2.match(thisLine)        
                if smtpLineDSN2Entries is not None:
                        queueid   = smtpLineDSN2Entries.group('queueid')
                        recip     = smtpLineDSN2Entries.group('recip')
                        orig      = smtpLineDSN2Entries.group('orig')

                        recipDomain = recip.lower().split('@')[-1]
		        
                        if queueid in self.surveyQueuid:
				if queueid in self.data.keys():
					if 'recipType' in self.data[queueid].keys():
						pass
					else:
						self.data[queueid]['recipType'] = {'external': 0 , 'internal': 0 }


                                	if self.data[queueid]['fromDomain'] == recipDomain :
                                       		self.data[queueid]['recipType']['internal'] += 1
                                	else:
                                        	self.data[queueid]['recipType']['external'] += 1
                                
					self.data[queueid]['recip'].append(recip.lower())
                                        
                        
                        
                        
                # When delivery is deferred (ie mailbox is overquota)        
                smtpLineDSN4Entries   = self.smtpLineDSN4.match(thisLine)        
                if smtpLineDSN4Entries is not None:
                        queueid       = smtpLineDSN4Entries.group('queueid')
                        deferredRecip = smtpLineDSN4Entries.group('recip')
                        orig          = smtpLineDSN4Entries.group('orig')
                        
                        recipDomain = deferredRecip.lower().split('@')[-1]
                        
                        # 1rst time : create
                        if queueid in self.surveyQueuid:
				if queueid in self.data.keys():

					if 'recipType' in self.data[queueid].keys():
                                        	pass
                               	 	else:
                                        	self.data[queueid]['recipType'] = {'external': 0 , 'internal': 0 }


                                	if self.data[queueid]['fromDomain'] == recipDomain :
                                        	self.data[queueid]['recipType']['internal'] += 1
                                	else:
                                        	self.data[queueid]['recipType']['external'] += 1
                                		self.data[queueid]['deferredRecip'].append(deferredRecip.lower())
                      
 
                        # 2nd time : update
                        if queueid in self.updateSurveyQueuid:
                                self.updateData[queueid]['deferredRecip'].append(deferredRecip.lower())
                                        
                                        
                       
                # When delivery is bounced (ie the mailbox does not exist or the msg is older than 5 days)        
                smtpLineDSN5Entries   = self.smtpLineDSN5.match(thisLine)        
                if smtpLineDSN5Entries is not None:
                        queueid      = smtpLineDSN5Entries.group('queueid')
                        bouncedRecip = smtpLineDSN5Entries.group('recip')
                        orig         = smtpLineDSN5Entries.group('orig')

                                       
                        recipDomain = bouncedRecip.lower().split('@')[-1]
                        
                        # 1rst time : create
                        if queueid in self.surveyQueuid:
				if queueid in self.data.keys():
					if 'recipType' in self.data[queueid].keys():
                                        	pass
                                	else:
                                        	self.data[queueid]['recipType'] = {'external': 0 , 'internal': 0 }
                                

                               	 	if self.data[queueid]['fromDomain'] == recipDomain :
                                        	self.data[queueid]['recipType']['internal'] += 1
                                	else:
                                        	self.data[queueid]['recipType']['external'] += 1 
                                
                                	self.data[queueid]['bouncedRecip'].append(bouncedRecip.lower())
                                                
                        
                        # 2nd time : update
                        if queueid in self.updateSurveyQueuid:
                                self.updateData[queueid]['bouncedRecip'].append(bouncedRecip.lower())               
                            
                            

                # Because we end the mail bloc log, we can now put it in ELK       
                smtpRemovedEntries = self.smtpRemoved.match(thisLine)
                if smtpRemovedEntries is not None:
                        queueid   = smtpRemovedEntries.group('queueid')
                        
                        # 1rst time : create
                        if queueid in self.surveyQueuid:
                        	if self.action == "storeToELK": 
					requestToELK().sendData(self.data[queueid])
					
	                # In any case we delete all about this message to preserve memory                
			try:
				del self.data[queueid]
			except KeyError:                        
 				pass

                        # 2nd time : update
                        if queueid in self.updateSurveyQueuid:
                        	if self.action == "storeToELK": 
                                        
					if len(self.updateData[queueid]['bouncedRecip']) or len(self.updateData[queueid]['deferredRecip']): 
                                        	u = requestToELK()
                                        	time.sleep(self.timeInterval)
                                        	u.retrieveBmsearchId(self.updateData[queueid]['msgid'])
                                        	u.replaceDSNRecip(self.updateData[queueid])
					
	                # In any case we delete all about this message to preserve memory                
			try:
				
				del self.updateData[queueid]
			except KeyError:                        
 				pass
                        	
		
			# Delete some bad entries 
			for badKey in ('table hash','warning','statistics'):
				if badKey in  self.data[queueid]:
		                	try:
                                		del self.data[queueid]
                        		except KeyError:
                                		pass

		

	def getData(self):
                """
                getData :
                        Retrieve data dict values.
                        Only for debug because must comment "del self.data[queueid]" line in parseContent() method.
                """
                
		return self.data
		
		
		

