#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 * ------------------------------------------------------------------------
 * bmSearch
 * ------------------------------------------------------------------------
 * @license : GNU Affero General Public License Version 3 <http://www.gnu.org/licenses>
 * @author  : Pascal SALAUN 
 * @Website : http://www.bm-stats.org <http://www.bm-stats.org>
 * ------------------------------------------------------------------------
"""

# Native Python Lib/Class 
import os
import sqlite3
import sys

# bmSearch Class


class sqlQuery:
        """
        sqlQuery : 
                This class offers all methods to manage message-id in time in datastoring them in a simple table index containing two columns ('messageid', 'timestamp') 
                This DB contain tables :
                        `index` (`messageid` INTEGER NOT NULL UNIQUE, `timestamp` INTEGER NOT NULL, PRIMARY KEY(messageid) )
        
        
        """


        def __init__(self, db):
                """
                 __init__ 
                        init the connection to the database stored in db file
                        Args :
                                "db" : the full path to the db file, by default '/var/lib/bmSearch/bmSearch.sqlite' 
                        
                """
                
                try:
                	self.conn = sqlite3.connect(db)
                	self.c = self.conn.cursor()
                	
        	except Exception as error:
    			pass



        def close(self):
                """
                close :
                        close the connection to the dabase
                        
                """
                try:
		        self.c.close()
		        del self.c
		        self.conn.close()
		        
                except Exception as error:
    			pass     
    			          


        def verifyIfMessageidExists(self, messageid):
                """
                verifyIfMessageidExists:
                        return "True" or "False" if tupl(messageid) exists or not in table "index"
                
                """
		query = "SELECT count(*) FROM `index` WHERE messageid=\"" + messageid + "\";"
		
		try:
		        self.c.execute(query)
		        res = self.c.fetchone()

		        if res[0]>0:
		                return True
		        else:
		                return False
		                
                except Exception as error:
    			pass


                
        def initDb(self):
                """
                initDb: 
                        create the sqlite database, by default '/var/lib/bmSearch/bmSearch.sqlite' 
                """
                try:
		        self.c.execute("""
		        CREATE TABLE IF NOT EXISTS `index` (
		                `messageid`	TEXT NOT NULL UNIQUE,
		                `timestamp`	INTEGER NOT NULL,
		                PRIMARY KEY(messageid)
		        );
		        """)
		        
		        self.c.execute("vacuum")
		        self.conn.commit()
		        
        	except Exception as error:
    			pass
        
        
        
        def updateTrigger(self, dbStoreMaxDays):
                """
                updateTrigger:
                        the bmSearch sqlite database has a trigger which delete entries older than dbStoreMaxDays days, 5 by default.
                        this method renew the trigger with new value
                """
		dropTrigger="""
		DROP TRIGGER IF EXISTS purge_old_data ;
		"""
		

		addTrigger="""
		CREATE TRIGGER IF NOT EXISTS purge_old_data AFTER INSERT ON `index`
			BEGIN
				DELETE FROM `index` WHERE timestamp < (strftime('%s', 'now') - """ + str(dbStoreMaxDays) + """*86400) ;
			END;
		"""
		
		try:
			self.c.execute(dropTrigger)
			self.c.execute(addTrigger)
			self.conn.commit()   
			     
         	except Exception as error:
    			pass
       
        
        
        def purgeIndexTable(self):
                """
                purgeIndexTable:
                        do a truncate of index table.
                """
                try:
                	self.c.execute("""truncate index ;""")
                	self.c.execute("vacuum")
                	self.conn.commit()
                
        	except Exception as error:
    			pass
                               
                
                
        def deleteEntriesOlderThan(self, tupl):
                """
                deleteEntriesOlderThan:
                        delete all index entries older than tupl(timestamp) value
                
                """
                try:
                        
                        self.c.execute("""delete FROM index WHERE timestamp<?""",tupl)
                        self.conn.commit()
                        return 0
                        
        	except Exception as error:
    			pass
        
        
        
        def deleteEntry(self, tupl):
                """
                deleteEntry:
                        delete single entry tupl(columnName, value) in index table.
                
                """
                try:                        
                        self.c.execute("""delete FROM index WHERE ?=?""",tupl)
                        self.conn.commit()
                        return 0
                        
        	except Exception as error:
    			pass
                
                

        def addEntry(self, tupl):
                """
                addEntry:
                        add in index table single tupl(messageid, timestamp)
                """
                query = "insert into `index` (messageid,timestamp) values (?,?) ;"
                
		try:
			self.c.execute(query, tupl)
			self.conn.commit()
			
        	except Exception as error:
    			pass
        
